from redis import Redis
from typing import Any, List, Optional, cast
from ...ioc.singleton import SingletonMeta


class SyncRedisClient(metaclass=SingletonMeta):
    def __init__(
        self,
        redis_connection: Optional[Redis] = None,
    ):
        if not hasattr(self, "redis") and isinstance(redis_connection, Redis):
            self.redis = redis_connection
        self.pubsub = self.redis.pubsub()

    def set(self, key: str, value: Any, expiry: Optional[int] = None) -> bool:
        return bool(self.redis.set(key, value, ex=expiry))

    def get(self, key: str) -> Optional[bytes]:
        return cast(Optional[bytes], self.redis.get(key))

    def delete(self, key: str) -> int:
        return cast(int, self.redis.delete(key))

    def delete_all_keys(self, keys: List[str]) -> int:
        return cast(int, self.redis.delete(*keys))

    def get_keys(self, pattern: str = "*") -> List[str]:
        return cast(List[str], self.redis.keys(pattern))

    def send_command(self, *commands: str) -> Any:
        return self.redis.execute_command(*commands)

    def close_connection(self):
        self.redis.close()



from abc import ABC, abstractmethod
import asyncio
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import json
from typing import Any, Dict, Generic, Type, TypeVar, Union
from fastapi import FastAPI
from pydantic import BaseModel, ValidationError

from ..utils.exceptions import JwtValidationError, UserRateLimitReachedError
from ..interfaces.interfaces_th import (
    Headers_TH,
    SQSClientCallBackResponse_TH,
)
from ..middlewares.message_middlewares import (
    MessageMiddleware,
    ExceptionLogger,
)

T = TypeVar("T", bound=BaseModel)


class AsyncTaskHandler(ABC, Generic[T]):
    pydantic_model_class: Type[T]

    def __init__(self, app: FastAPI) -> None:
        self.app = app

    @classmethod
    @abstractmethod
    def execute_business_logic(cls, payload: T) -> None:
        """Execute the specific business logic for the action."""
        raise NotImplementedError("Subclasses must implement execute_business_logic")

    @classmethod
    def context_setter_and_execute_payload(
        cls, payload: Dict[str, Any], headers: Headers_TH
    ) -> None:
        """Set the context and execute the business logic."""
        MessageMiddleware.validate_set_context(
            payload=payload,
            headers=headers,
            payload_pydantic_model=cls.pydantic_model_class,
        )
        cls.execute_business_logic(payload=cls.pydantic_model_class(**payload))

    async def _handle_with_executor(
        self,
        payload: Dict[str, Any],
        headers: Headers_TH,
        executor: Union[ProcessPoolExecutor, ThreadPoolExecutor],
    ) -> str:
        """Common handler logic for both process and thread pools."""
        correlation_id = headers["correlationid"]
        try:
            future = executor.submit(
                self.context_setter_and_execute_payload,
                payload=payload,
                headers=headers,
            )
            await asyncio.wrap_future(future)
            success = True
        except (JwtValidationError, ValidationError, UserRateLimitReachedError):
            # In case of JwtValidationError ,Pydantic ValidationError and  UserRateLimitReached Error   we  delete the message from the queue .
            ExceptionLogger.log_exception(self, correlation_id=correlation_id)
            success = True
        except Exception:
            # For other exceptions, we consider the task failed and do not  delte the message from  the queue . Note  : It will keep on retrying  .
            ExceptionLogger.log_exception(self, correlation_id=correlation_id)
            success = False
        response = SQSClientCallBackResponse_TH(
            allSuccess=success, correlationid=correlation_id
        )
        return json.dumps(response)

    async def handle_with_process_pool(
        self, payload: Dict[str, Any], headers: Headers_TH
    ) -> str:
        """Handle the task using a process pool executor."""
        process_pool: ProcessPoolExecutor = self.app.state.process_pool
        return await self._handle_with_executor(payload, headers, process_pool)

    async def handle_with_thread_pool(
        self, payload: Dict[str, Any], headers: Headers_TH
    ) -> str:
        """Handle the task using a thread pool executor."""
        thread_pool: ThreadPoolExecutor = self.app.state.thread_pool
        return await self._handle_with_executor(payload, headers, thread_pool)



from ast import List
from functools import lru_cache
import json
from jwcrypto import jwk
import jwt
from typing import Dict, List, Optional

from ..constants.errors import ErrorEnums

from ..utils.env_initlializer import EnvStore
from ..utils.exceptions import JwtValidationError
from ..interfaces.interfaces_th import Jwk_TH


class JwtValdationUtils:
    JWT_ALGORITHM = "RS256" if EnvStore().auth_token_algorithm == "RS256" else "HS256"

    @staticmethod
    def is_token_expired(token: str) -> bool:
        try:
            JwtValdationUtils.validate_token(token, verify_exp=True)
            return False
        except JwtValidationError as e:
            if e == ErrorEnums.TOKEN_EXPIRED:
                return True
            raise e

    @lru_cache(maxsize=1)
    @staticmethod
    def __load_jwks() -> List[Jwk_TH]:
        return json.loads(EnvStore().jwks)

    @classmethod
    def _get_public_key(cls, kid: str) -> Optional[str]:
        public_key = next((key for key in cls.__load_jwks() if key["kid"] == kid), None)
        if public_key:
            return jwk.JWK(**public_key).export_to_pem().decode("utf-8")
        return None

    @classmethod
    def validate_token(
        cls,
        token: str,
        verify_exp: bool = False,
        verify_aud: bool = False,
    ) -> Dict:
        try:
            JWT_ALGORITHM = (
                "RS256" if EnvStore().auth_token_algorithm == "RS256" else "HS256"
            )
            default_options = dict()
            if not verify_exp:
                default_options.update(verify_exp=False)
            if not verify_aud:
                default_options.update(verify_aud=False)

            unverified_header = jwt.get_unverified_header(token)
            public_key = cls._get_public_key(unverified_header["kid"])

            if not public_key:
                raise ValueError("No matching public key found")

            return jwt.decode(
                token,
                public_key,
                algorithms=[JWT_ALGORITHM],
                options=default_options,
            )
        except jwt.ExpiredSignatureError:
            raise JwtValidationError(ErrorEnums.TOKEN_EXPIRED)
        except jwt.InvalidSignatureError:
            raise JwtValidationError(ErrorEnums.INVALID_SIGNATURE)
        except jwt.InvalidAudienceError:
            raise JwtValidationError(ErrorEnums.INVALID_AUDIENCE)
        except jwt.InvalidIssuerError:
            raise JwtValidationError(ErrorEnums.INVALID_ISSUER)
        except jwt.DecodeError:
            raise JwtValidationError(ErrorEnums.MALFORMED_TOKEN)
        except jwt.PyJWTError as e:
            raise JwtValidationError(f"Invalid token: {str(e)}")
        except Exception:
            raise Exception(ErrorEnums.JWT_GENERAL_ERROR)



from enum import StrEnum


class ErrorEnums(StrEnum):
    TOKEN_EXPIRED = "Token has expired"
    INVALID_SIGNATURE = "Invalid token: Signature verification failed"
    INVALID_AUDIENCE = "Invalid token: Audience claim verification failed"
    INVALID_ISSUER = "Invalid token: Issuer claim verification failed"
    MALFORMED_TOKEN = "Malformed token: Unable to decode"
    JWT_GENERAL_ERROR = "Internal server error during token validation"



from typing import Optional, cast
from ..ioc.singleton import SingletonMeta


class EnvStore(metaclass=SingletonMeta):
    def __init__(self) -> None:
        # Service  Utils  required envs
        self._aws_region_name: Optional[str] = None
        self._aws_access_key_id: Optional[str] = None
        self._aws_secret_access_key: Optional[str] = None
        self._domain: Optional[str] = None
        self._servicename: Optional[str] = None
        self._environment: Optional[str] = None
        self._dbname: Optional[str] = None
        self._jwks: Optional[str] = None
        self._auth_token_algorithm: Optional[str] = None

    def validate_env_variables(self) -> None:
        missing_vars = []

        for attr in self.__dict__:
            if getattr(self, attr) is None:
                missing_vars.append(attr.upper())

        if missing_vars:
            raise ValueError(
                f"Missing environment variables: {', '.join(missing_vars)}"
            )

    @property
    def jwks(self) -> str:
        return cast(str, self._jwks)

    @jwks.setter
    def jwks(self, value: str) -> None:
        self._jwks = value

    @property
    def dbname(self) -> str:
        return cast(str, self._dbname)

    @dbname.setter
    def dbname(self, value: str) -> None:
        self._dbname = value

    @property
    def environment(self) -> str:
        return cast(str, self._environment)

    @environment.setter
    def environment(self, value: str) -> None:
        self._environment = value

    @property
    def servicename(self) -> str:
        return cast(str, self._servicename)

    @servicename.setter
    def servicename(self, value: str) -> None:
        self._servicename = value

    @property
    def domain(self) -> str:
        return cast(str, self._domain)

    @domain.setter
    def domain(self, value: str) -> None:
        self._domain = value

    @property
    def aws_region_name(self) -> str:
        return cast(str, self._aws_region_name)

    @aws_region_name.setter
    def aws_region_name(self, value: str) -> None:
        self._aws_region_name = value

    @property
    def aws_access_key_id(self) -> str:
        return cast(str, self._aws_access_key_id)

    @aws_access_key_id.setter
    def aws_access_key_id(self, value: str) -> None:
        self._aws_access_key_id = value

    @property
    def aws_secret_access_key(self) -> str:
        return cast(str, self._aws_secret_access_key)

    @aws_secret_access_key.setter
    def aws_secret_access_key(self, value: str) -> None:
        self._aws_secret_access_key = value

    @property
    def auth_token_algorithm(self) -> str:
        return cast(str, self._auth_token_algorithm)

    @auth_token_algorithm.setter
    def auth_token_algorithm(self, value: str) -> None:
        self._auth_token_algorithm = value



import boto3
import json
from typing import Optional, final, Dict
from typing import TypedDict, Required, Any

from ...utils.env_initlializer import EnvStore

from ...utils.context_utils import ContextUtils

from .helpers import Helpers
from .queue_config import QUEUE_CONFIG
from ...logging.base_logger import APP_LOGGER


class Message_Payload_TypeHinter(TypedDict, total=False):
    action: Required[str]
    payload: Dict[str, Any]
    context: Dict[str, Any]


@final
class SyncSQSPusher:
    queue_config: Dict[str, Dict[str, bool]] = QUEUE_CONFIG

    def __construct_message_body(
        self, action: str, payload: Dict[str, Any]
    ) -> Message_Payload_TypeHinter:
        message_context = ContextUtils.get_header_details().model_dump(mode="json")
        message_payload = Message_Payload_TypeHinter(
            action=action, context=message_context, payload=payload
        )
        return message_payload

    def __init__(self, queue_name: str, action: str, payload: Dict[str, Any]) -> None:
        self.sqs_client = boto3.client(
            "sqs",
            aws_access_key_id=EnvStore().aws_access_key_id,
            aws_secret_access_key=EnvStore().aws_secret_access_key,
            region_name=EnvStore().aws_region_name,
        )
        self.aws_account_id = boto3.client(
            "sts",
            aws_access_key_id=EnvStore().aws_access_key_id,
            aws_secret_access_key=EnvStore().aws_secret_access_key,
        ).get_caller_identity()["Account"]
        self.queue_name = queue_name
        self.queue_url = Helpers.construct_queue_url(
            queue_name=queue_name,
            region_name=EnvStore().aws_region_name,
            params=self.queue_config[queue_name],
            aws_account_id=self.aws_account_id,
        )
        self.message_payload = self.__construct_message_body(
            action=action, payload=payload
        )

    def post_message(self) -> None:
        try:
            response = self.sqs_client.send_message(
                QueueUrl=self.queue_url, MessageBody=json.dumps(self.message_payload)
            )
            APP_LOGGER.info(
                f"Message sent successfully to queue '{self.queue_name}' with Message ID: { response['MessageId']} and  ->  message payload : {self.message_payload}"
            )
        except Exception as e:
            APP_LOGGER.error("Error sending message:" + str(e))
            raise e

    def post_message_to_fifo(
        self, message_group_id: str, message_deduplication_id: Optional[str] = None
    ) -> None:
        """
        Post a message to a FIFO queue.

        Args:
            message_group_id (str): The tag that specifies that a message belongs to a specific message group.
                                   Messages that belong to the same message group are processed in FIFO order.
            message_deduplication_id (str, optional): The token used for deduplication of sent messages.
                                                     If not provided, content-based deduplication will be used
                                                     (if enabled for the queue).

        Returns:
            None

        Raises:
            Exception: If there's an error sending the message.
        """
        try:
            if not self.queue_name.endswith(".fifo"):
                APP_LOGGER.warning(
                    f"Queue '{self.queue_name}' does not appear to be a FIFO queue (name should end with .fifo)"
                )
                return
            message_params = {
                "QueueUrl": self.queue_url,
                "MessageBody": json.dumps(self.message_payload),
                "MessageGroupId": message_group_id,
            }

            if message_deduplication_id:
                message_params["MessageDeduplicationId"] = message_deduplication_id
            response = self.sqs_client.send_message(**message_params)

            APP_LOGGER.info(
                f"Message sent successfully to FIFO queue '{self.queue_name}' with Message ID: {response['MessageId']}, "
                f"MessageGroupId: {message_group_id} and message payload: {self.message_payload}"
            )
        except Exception as e:
            APP_LOGGER.error(f"Error sending message to FIFO queue: {str(e)}")
            raise e



from fastapi import HTTPException
from typing import cast
from redis.exceptions import RedisError

from ..utils.exceptions import UserRateLimitReachedError

from .guard_key_enums import RateLimiterGuardKeys

from ..clients.redis_client.async_redisclient import (
    AsyncRedisClient,
)

from ..clients.redis_client.enums import RedisExpiryEnums


class RateLimiterGuard:
    def __init__(
        self,
        key: RateLimiterGuardKeys,
        cache_expiry: RedisExpiryEnums,
        max_calls: int,
        raiseHttpError: bool = True,
    ):
        self.__redis_client = AsyncRedisClient()
        self.__key = key
        self.__cache_expiry = cache_expiry
        self.__max_calls = max_calls
        self.__raiseHttpError = raiseHttpError

    @staticmethod
    async def __implement(
        redis_client: AsyncRedisClient,
        key: str,
        cache_expiry: RedisExpiryEnums,
        max_calls: int,
        raiseHttpError: bool,
    ) -> None:
        """Checks if the request exceeds the rate limit."""
        try:
            match cache_expiry:
                case RedisExpiryEnums.ONE_MIN_EXPIRY:
                    cache_key = f"{key}:ONE_MIN_EXPIRY"
                case RedisExpiryEnums.ONE_DAY_EXPIRY:
                    cache_key = f"{key}:ONE_DAY_EXPIRY"
                case RedisExpiryEnums.ONE_HOUR_EXPIRY:
                    cache_key = f"{key}:ONE_HOUR_EXPIRY"
                case RedisExpiryEnums.ONE_MONTH_EXPIRY:
                    cache_key = f"{key}:ONE_MONTH_EXPIRY"
                case _:
                    raise ValueError("Invalid Redis Key.")

            current_count = cast(
                int, await redis_client.send_command("INCR", cache_key)
            )
            if current_count == 1:
                await redis_client.send_command("EXPIRE", cache_key, cache_expiry.value)
            if current_count > max_calls:
                if raiseHttpError:
                    raise HTTPException(
                        status_code=429,
                        detail=f"Rate limit exceeded. Max {max_calls} requests allowed.",
                    )
                else:
                    raise UserRateLimitReachedError(
                        f"Rate limit exceeded. Max {max_calls} requests allowed."
                    )
        except RedisError:
            raise Exception("Rate limiter failed due to Redis error.")

    async def __call__(self) -> None:
        await self.__implement(
            redis_client=self.__redis_client,
            key=self.__key,
            cache_expiry=self.__cache_expiry,
            max_calls=self.__max_calls,
            raiseHttpError=self.__raiseHttpError,
        )



from enum import StrEnum


class RateLimiterGuardKeys(StrEnum):
    BEDROCK_LLM_CALL_GUARD = "BEDROCK_LLM_CALL_GUARD"
    BEDROCK_EMBEDDING_CALL_GUARD = "BEDROCK_EMBEDDING_CALL_GUARD"
    COGNITO_GUARD = "COGNITO_GURAD"
    TEST_KEY = "TEST_KEY"
